// JISToUNI2.h | by Doug Manley
///////////////////////////////////////////////////////////
//    The purpose of this file is to provide the
// ConvertEncoding() functions for converting any type of
// encoding to any other type.
///////////////////////////////////////////////////////////
#ifndef JISTOUNI2_H
#define JISTOUNI2_H
///////////////////////////////////////////////////////////
#include <GNU\iconv.h>
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
/*		This converts text ('text') from one encoding ('from')
	to another encoding ('to').
		It is the responsibility of the caller to deal with the
	char* pointer that ConvertEncoding() returns.  Since it was
	created with 'new', it MUST be 'delete'd at some point.
*/
char* ConvertEncoding( char *text,int text_size, char *from,char *to ) {
	// Set it so that there's no error...
	// (because iconv doesn't seem to do this on its own)...
	errno = 0;

	// Open the descriptor for the conversion:
	iconv_t descriptor = iconv_open( to, from );

	if( descriptor == (iconv_t)-1 ) {
		if( errno == EINVAL ) {
			Log(L"ERROR: ConvertEncoding(): errno == EINVAL.\n");
			TCHAR cstr[1024];
				swprintf(cstr,L"conversion from '%s' to '%s' not available", from, to );
				wcscat( cstr,L"\n\nPlease report this error to Doug Manley (ssvegeta7@yahoo.com), please." );
			MessageBox(NULL,cstr,TEXT("Error: iconv_open"), MB_OK );
		} else {
			Log(L"ERROR: ConvertEncoding(): errno == (something that is not EINVAL).\n");
			TCHAR cstr[1024];
				swprintf(cstr,L"errno: %i\n\nPlease report this error to Doug Manley (ssvegeta7@yahoo.com), please.", errno );
			MessageBox(NULL,cstr,TEXT("Error: iconv_open"), MB_OK );
		}
		return(NULL);
	}

	// Okay, now that we have the descript open,
	// We need to set up our input & output buffers.

	// The input buffer is just the text that we were given:
	char *in_buffer = text;
	// The output buffer is what it writes the converted string to.
	// Now, since it COULD be in Unicode (32-bit), we make the buffer
	// four times as big as we would normally.
	int out_buffer_size = 4*(text_size+1);
	char *out_buffer = new char[out_buffer_size];
	if( !out_buffer ) {
		Log(L"ERROR: ConvertEncoding(): out_buffer not created.\n");
		MessageBox(NULL,L"out_buffer not created!",L"Error",MB_OK );
		return(NULL);
	}
	// Clear the output buffer:
	strnset( out_buffer, '\0', out_buffer_size );


	// Now, it's actually time to convert the text.

	// 'inbytes' is the size of the text that we're putting in.
	// 'outbytes' is the size of the text that we can get out of this.
	size_t inbytes = text_size,
   		 outbytes = out_buffer_size - 4; // 4 is for the 32-bit padding.


	// NOTE:
	//    The following "make-some-extra-pointers" method WORKS.
	//    --From what I can tell, it's a C thing :(
	// DO NOT MESS WITH IT.
	// (You will only screw things up; trust me.)

	// Make some new pointers:
	char *in = in_buffer;
	char *out = out_buffer;

	// CONVERSION TIME:
	size_t nconv = iconv( descriptor, (const char**)&in, &inbytes,
																   &out,&outbytes
	);
	// (on error, this returns the number of bytes left to convert).

	// NOW, outbytes is the number of FREE BYTES in outbuffer.
	// (minus 4, since we held back).


	/* 2005.01.01: I know that it sounds stupid, but this is fairly important
	               to have, so just keep it. */
	// Fill in the rest of the buffer... to play it safe.
	for( int i = 0; i < outbytes+4; i++ ) {
		// Bugfix: 2004.12.14, by "Truong Hoang Nguyen" <thnguyen@shinyoko.isb.co.jp>.
		// (the '-1' is very important for preventing array overflowing).
		out_buffer[(out_buffer_size-1)-i] = '\0';
	}//*/

	if( debug_level >= 2 ) {
		Log(2,L"*****************\n");
		Log(2,L"***   ICONV   ***\n");
		Log(2,L"*****************\n");
		Log(2,L"out_buffer: size (original): %i\n", out_buffer_size-4 );
		Log(2,L"out_buffer: size (outbytes): %i\n", (out_buffer_size-4)-outbytes );
		Log(2,L"out_buffer:   <%s>\n", out_buffer );//*/
	}

	// I forget what this signifies; I believe that 'inbytes' is the number
	// of bytes that iconv did NOT get to process...
	if( inbytes != 0 ) {
		Log(L"ERROR: ConvertEncoding(): Iconv: inbytes is not zero\n");
	}

	// Close the descriptor:
	iconv_close( descriptor );

	if( nconv == (size_t)0 ) {
		// SUCCESS!
		return( out_buffer );
	} else if( nconv == (size_t)(-1) ) {
	// It's an error if there are still bytes to convert.
		if( errno == E2BIG ) {
			Log(L"ERROR: ConvertEncoding(): nconv == -1 | errno == E2BIG.\n");
			MessageBox(NULL,TEXT("There is not sufficient room at *outbuf\n\nPlease report this error to Doug Manley (ssvegeta7@yahoo.com), please."),TEXT("Error: iconv"), MB_OK );
		} else if( errno == EILSEQ	) {
			Log(L"ERROR: ConvertEncoding(): nconv == -1 | errno == EILSEQ.\n");
			MessageBox(NULL,TEXT("An invalid multibyte sequence has been encountered in the input."),TEXT("Error: iconv"), MB_OK );
		} else if( errno == EINVAL	) {
			Log(L"ERROR: ConvertEncoding(): nconv == -1 | errno == INVAL.\n");
			MessageBox(NULL,TEXT("An incomplete multibyte sequence has been encountered in the input."),TEXT("Error: iconv"),MB_OK );
		} else if( errno == 0 ) {
		// errno == 0 seems to be when it just can't convert the text (not right encoding).
			Log(L"ERROR: ConvertEncoding(): nconv == -1 | errno == 0 | out_buffer: (%S).\n",out_buffer );
		} else {
			Log(L"ERROR: ConvertEncoding(): nconv == -1 | errno == (nothing good: %i).\n", errno );
			TCHAR cstr[1024];
				swprintf(cstr,L"errno: %i\n\nPlease report this error to Doug Manley (ssvegeta7@yahoo.com), please.", errno );
			MessageBox(NULL,cstr,TEXT("Error: iconv"), MB_OK );
		}
	} else {
	// The result is the number of non-reversible conversions made.
	}

	// We failed if we got to this point, so clean up:
	delete [] out_buffer;
	return(NULL);
}

int ConvertEncoding( char *text,int text_size, char *from,char *to, wstring &strline ) {
	char *result = ConvertEncoding( text,text_size, from,to );
	if( result ) {
		// Set the string to be our result.
		strline = (Uint16*)result;
		// And... delete the pointer.
		delete [] result;
		return(0);
	} else {
		return(-1);
	}
}

int ConvertEncoding( char *text,int text_size, char *from,char *to, string &strline ) {
	char *result = ConvertEncoding( text,text_size, from,to );
	if( result ) {
		// Set the string to be our result.
		strline = (char*)result;
		// And... delete the pointer.
		delete [] result;
		return(0);
	} else {
		return(-1);
	}
}
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
#endif
///////////////////////////////////////////////////////////
